//
//  FeedListViewModel.swift
//  ReviewsFeed
//
//  Created by Javier Cicchelli on 18/03/2024.
//  Copyright © 2024 Röck+Cöde. All rights reserved.
//

import Foundation
import ReviewsFilterKit
import ReviewsiTunesKit

extension FeedListViewController {
    final class ViewModel: ObservableObject {
        
        // MARK: Constants
        private let configuration: FeedListConfiguration
        
        private let filterWords: FilterWordsUseCase = .init()
        private let topWords: TopWordsUseCase = .init()

        // MARK: Properties
        @Published private(set) var filter: FilterOption = .all
        @Published private(set) var isFilterEnabled: Bool = false
        @Published private(set) var isFiltering: Bool = false
        @Published private(set) var isLoading: Bool = false
        @Published private(set) var state: FeedListState = .initial

        private(set) var items: [Review] = []
        private(set) var words: [TopWord] = []

        private var reviewsAll: [Review] = []
        private var reviewsFiltered: FilteredReviews = [:]
        private var reviewsTopWords: TopWordsReviews = [:]
        
        private weak var coordination: FeedListCoordination?
        
        lazy private var iTunesService: iTunesService = {
            .init(configuration: .init(session: configuration.session))
        }()

        // MARK: Initialisers
        init(
            configuration: FeedListConfiguration = .init(),
            coordination: FeedListCoordination? = nil
        ) {
            self.configuration = configuration
            self.coordination = coordination
        }
        
        // MARK: Computed
        var itemsCount: Int {
            isWordsShowing
                ? items.count + 1
                : items.count
        }
        
        var isWordsShowing: Bool {
            filter != .all
                && !words.isEmpty
        }

        // MARK: Functions
        func fetch() async {
            isFilterEnabled = false
            isLoading = items.isEmpty

            do {
                let output = try await iTunesService.getReviews(.init(
                    appID: configuration.appID,
                    countryCode: configuration.countryCode
                ))
                
                reviewsAll = output.reviews.map(Review.init)
                reviewsFiltered = FilterOption.allCases
                    .reduce(into: FilteredReviews()) { partialResult, option in
                        partialResult[option] = reviewsAll.filter { $0.rating.stars == option.rawValue }
                    }
                reviewsTopWords = reviewsFiltered
                    .mapValues { reviews in
                        reviews.map(\.comment)
                            .compactMap { try? filterWords($0) }
                    }
                    .mapValues {
                        topWords($0).map(TopWord.init)
                    }

                items = filter == .all
                    ? reviewsAll
                    : reviewsFiltered[filter] ?? []
                words = filter == .all
                    ? []
                    : reviewsTopWords[filter] ?? []

                isFilterEnabled = !items.isEmpty
                state = items.isEmpty
                    ? .empty
                    : .populated
            } catch {
                items = []
                state = .error
            }

            isLoading = false
        }
        
        func filter(by option: FilterOption) {
            guard option != filter else { return }
            
            items = option == .all
                ? reviewsAll
                : reviewsFiltered[option] ?? []
            words = reviewsTopWords[option] ?? []

            filter = option
        }
        
        func item(for index: Int) -> Review? {
            guard !items.isEmpty else {
                return nil
            }
            
            let indexToUse = isWordsShowing
                ? index - 1
                : index
            
            guard indexToUse < items.count else {
                return nil
            }
            
            return items[indexToUse]
        }
        
        func openItem(at index: Int) {
            guard let item = item(for: index) else { return }
            
            coordination?.open(item)
        }

    }
}

// MARK: - Helpers
private extension FeedListViewController.ViewModel {
    
    // MARK: Type aliases
    typealias FilteredReviews = [FilterOption: [Review]]
    typealias TopWordsReviews = [FilterOption: [TopWord]]

}
