//
//  FeedListViewController.swift
//  ReviewsFeed
//
//  Created by Dmitrii Ivanov on 21/07/2020.
//  Copyright © 2020 ING. All rights reserved.
//

import Combine
import Foundation
import ReviewsFoundationKit
import ReviewsUIKit
import SwiftUI
import UIKit

final class FeedListViewController: UIViewController {
    
    // MARK: Constants
    private let viewModel: ViewModel
    
    // MARK: Properties
    private var unavailableView: FeedUnavailableView?
    
    private var cancellables: Set<AnyCancellable> = []
    
    // MARK: Outlets
    private lazy var activityIndicator = {
        let indicator = UIActivityIndicatorView()
        
        indicator.translatesAutoresizingMaskIntoConstraints = false
        
        return indicator
    }()
    
    private lazy var filterButton = {
        UIBarButtonItem(
            title: NSLocalizedString(
                .Key.Navigation.Button.filter,
                bundle: .module,
                comment: .empty
            ),
            image: .Icon.filter,
            primaryAction: nil,
            menu: .init(
                title: NSLocalizedString(
                    .Key.Menu.filter,
                    bundle: .module,
                    comment: .empty
                ),
                image: UIImage.Icon.star,
                children: FilterOption.allCases.map { option -> UIAction in
                        .init(
                            title: option.text,
                            image: .init(systemName: option.icon)
                        ) { [weak self] _ in
                            self?.viewModel.filter(by: option)
                        }
                }
            )
        )
    }()
    
    private lazy var pullControl = {
        let control = UIRefreshControl()
        
        control.attributedTitle = .init(string: NSLocalizedString(
            .Key.PullToRefresh.title,
            bundle: .module,
            comment: .empty
        ))
        
        control.addTarget(
            self,
            action: #selector(refresh(_:)),
            for: .valueChanged
        )
        
        return control
    }()
    
    private lazy var tableView = {
        let table = UITableView(
            frame: .zero,
            style: .plain
        )

        table.translatesAutoresizingMaskIntoConstraints = false

        return table
    }()
    
    // MARK: Initialisers
    init(_ viewModel: ViewModel) {
        self.viewModel = viewModel
        
        super.init(
            nibName: nil,
            bundle: nil
        )
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    // MARK: Computed
    var items: [Review] {
        viewModel.items
    }
    
    // MARK: UIViewController
    override func viewDidLoad() {
        super.viewDidLoad()
        
        setNavigationBar()
        setView()
        setLayout()
        registerTableCells()
        bindViewModel()
        
        Task { await viewModel.fetch() }
    }
    
}

// MARK: - UITableViewDataSource
extension FeedListViewController: UITableViewDataSource {
    
    // MARK: Functions
    func tableView(
        _ tableView: UITableView,
        numberOfRowsInSection section: Int
    ) -> Int {
        viewModel.itemsCount
    }
    
    func tableView(
        _ tableView: UITableView,
        cellForRowAt indexPath: IndexPath
    ) -> UITableViewCell {
        if viewModel.isWordsShowing && indexPath.row == 0 {
            makeTopWordsCell(tableView)
        } else {
            makeFeedItemCell(tableView, at: indexPath.row)
        }
    }
    
}

// MARK: - UITableViewDelegate
extension FeedListViewController: UITableViewDelegate {

    // MARK: Functions
    func tableView(
        _ tableView: UITableView,
        didSelectRowAt indexPath: IndexPath
    ) {
        viewModel.openItem(at: indexPath.row)
        
        tableView.deselectRow(
            at: indexPath,
            animated: true
        )
    }
    
}

// MARK: - Helpers
private extension FeedListViewController {

    // MARK: Actions
    @objc func refresh(_ sender: AnyObject) {
        Task { await self.viewModel.fetch() }
    }
    
    // MARK: Functions
    func bindViewModel() {
        viewModel.$filter
            .receive(on: RunLoop.main)
            .sink { [weak self] option in
                self?.updateFilterMenu(option)
                self?.tableView.reloadData()
            }
            .store(in: &cancellables)

        viewModel.$isFilterEnabled
            .removeDuplicates()
            .receive(on: RunLoop.main)
            .sink { [weak self] enabled in
                self?.filterButton.isEnabled = enabled
            }
            .store(in: &cancellables)
        
        viewModel.$isLoading
            .filter { $0 == true }
            .receive(on: RunLoop.main)
            .sink { [weak self] _ in
                self?.activityIndicator.startAnimating()
            }
            .store(in: &cancellables)

        viewModel.$isLoading
            .dropFirst()
            .filter { $0 == false }
            .receive(on: RunLoop.main)
            .sink { [weak self] _ in
                self?.activityIndicator.stopAnimating()
                self?.pullControl.endRefreshing()
                self?.tableView.reloadData()

                if self?.viewModel.state == .populated {
                    self?.tableView.scrollToRow(
                        at: .init(row: 0, section: 0),
                        at: .middle,
                        animated: true
                    )
                }
            }
            .store(in: &cancellables)
        
        viewModel.$state
            .dropFirst()
            .receive(on: RunLoop.main)
            .sink { [weak self] state in
                switch state {
                case .empty,
                     .error:
                    self?.insertUnavailableView(state)
                default:
                    self?.removeUnavailableView()
                }
                
            }
            .store(in: &cancellables)
    }
    
    func insertUnavailableView(_ state: FeedListState) {
        guard unavailableView == nil else { return }
        
        let isErrorState = state == .error
        let unavailable = FeedUnavailableView(
            systemImage: isErrorState
                ? .Icon.error
                : .Icon.empty,
            title: NSLocalizedString(
                isErrorState
                    ? .Key.Unavailable.Title.error
                    : .Key.Unavailable.Title.empty,
                bundle: .module,
                comment: .empty
            ),
            description: NSLocalizedString(
                isErrorState
                    ? .Key.Unavailable.Description.error
                    : .Key.Unavailable.Description.empty,
                bundle: .module,
                comment: .empty
            ),
            button: isErrorState
                ? NSLocalizedString(
                    .Key.Unavailable.Button.tryAgain,
                    bundle: .module,
                    comment: .empty
                )
                : nil,
            action: isErrorState
                ? { Task { await self.viewModel.fetch() } }
                : nil
        )
        
        addChild(unavailable)

        view.addSubview(unavailable.view)
        view.bringSubviewToFront(unavailable.view)
        
        // TODO: fix [this issue](https://stackoverflow.com/questions/70032739/why-does-swiftui-uihostingcontroller-have-extra-spacing) so autolayout can be used instead.
        unavailable.view.frame = tableView.bounds

        unavailable.didMove(toParent: self)

        unavailableView = unavailable
    }
    
    func makeFeedItemCell(
        _ tableView: UITableView,
        at row: Int
    ) -> UITableViewCell {
        guard
            let cell = FeedItemCell.dequeue(from: tableView),
            let item = viewModel.item(for: row)
        else {
            return .init()
        }
        
        cell.separatorInset = .init(width: tableView.bounds.size.width)
        cell.contentConfiguration = {
            if #available(iOS 16.0, *) {
                UIHostingConfiguration {
                    FeedItemView(item)
                }
            } else {
                HostingConfiguration {
                    FeedItemView(item)
                }
            }
        }()
        
        return cell
    }
    
    func makeTopWordsCell(_ tableView: UITableView) -> UITableViewCell {
        guard let cell = TopWordsCell.dequeue(from: tableView) else {
            return .init()
        }
        
        cell.separatorInset = .init(width: tableView.bounds.size.width)
        cell.contentConfiguration = {
            if #available(iOS 16.0, *) {
                UIHostingConfiguration {
                    TopWordsView(viewModel.words)
                }
            } else {
                HostingConfiguration {
                    TopWordsView(viewModel.words)
                }
            }
        }()

        return cell
    }

    func registerTableCells() {
        FeedItemCell.register(in: tableView)
        TopWordsCell.register(in: tableView)
    }
    
    func removeUnavailableView() {
        guard let unavailableView else { return }
        
        unavailableView.view.removeFromSuperview()

        self.unavailableView = nil
    }

    func setLayout() {
        NSLayoutConstraint.activate([
            activityIndicator.centerXAnchor.constraint(equalTo: view.centerXAnchor),
            activityIndicator.centerYAnchor.constraint(equalTo: view.safeAreaLayoutGuide.centerYAnchor),
            tableView.bottomAnchor.constraint(equalTo: view.bottomAnchor),
            tableView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
            tableView.topAnchor.constraint(equalTo: view.topAnchor),
            tableView.trailingAnchor.constraint(equalTo: view.trailingAnchor),
        ])
    }
    
    func setNavigationBar() {
        navigationController?.navigationBar.prefersLargeTitles = true
        navigationController?.navigationBar.isTranslucent = true

        navigationItem.rightBarButtonItem = filterButton
        navigationItem.title = NSLocalizedString(
            .Key.Navigation.title,
            bundle: .module,
            comment: .empty
        )
    }
    
    func setView() {
        view.addSubview(tableView)
        view.addSubview(activityIndicator)
        view.bringSubviewToFront(activityIndicator)
        
        tableView.delegate = self
        tableView.refreshControl = pullControl
        tableView.dataSource = self

        activityIndicator.startAnimating()
    }
    
    func updateFilterMenu(_ option: FilterOption) {
        filterButton
            .menu?
            .children
            .compactMap { $0 as? UIAction }
            .forEach { action in
                action.state = action.title == option.text
                    ? .on
                    : .off
            }
    }
    
}

// MARK: - String+Constants
private extension String {
    enum Cell {
        static let feedItem = "FeedItemCell"
    }
    
    enum Key {
        enum Menu {
            static let filter = "common.filter.menu.title.text"
        }

        enum Navigation {
            static let title = "view.feed-list.navigation-bar.title.text"
            
            enum Button {
                static let filter = "view.feed-list.navigation-bar.button.filter-list.text"
            }
        }
        
        enum PullToRefresh {
            static let title = "view.feed-list.pull-to-refresh.title.text"
        }
        
        enum Unavailable {
            enum Button {
                static let tryAgain = "view.feed-list.unavailable.error.button.try-again.text"
            }
            
            enum Description {
                static let empty = "view.feed-list.unavailable.empty.description.text"
                static let error = "view.feed-list.unavailable.empty.description.text"
            }
            
            enum Title {
                static let empty = "view.feed-list.unavailable.empty.title.text"
                static let error = "view.feed-list.unavailable.error.title.text"
            }
        }
    }
}

// MARK: - UIEdgeInsets+Inits
private extension UIEdgeInsets {
    
    // MARK: Initialisers
    init(width: CGFloat) {
        self = .init(
            top: 0,
            left: width,
            bottom: 0,
            right: 0
        )
    }
    
}

// MARK: - Previews
#if DEBUG
import ReviewsiTunesKit

@available(iOS 17.0, *)
#Preview("Feed List with few reviews") {
    MockURLProtocol.response = .init(
        statusCode: 200,
        object: Feed(entries: .sample)
    )
    
    return UINavigationController(rootViewController: FeedListViewController(.init(
        configuration: .init(session: .mock)
    )))
}

@available(iOS 17.0, *)
#Preview("Feed List with no reviews") {
    MockURLProtocol.response = .init(
        statusCode: 200,
        object: Feed(entries: [])
    )
    
    return UINavigationController(rootViewController: FeedListViewController(.init(
        configuration: .init(session: .mock)
    )))
}

@available(iOS 17.0, *)
#Preview("Feed List with error when loading reviews") {
    MockURLProtocol.response = .init(statusCode: 404)
    
    return UINavigationController(rootViewController: FeedListViewController(.init(
        configuration: .init(session: .mock)
    )))
}

@available(iOS 17.0, *)
#Preview("Feed List with live reviews") {
    return UINavigationController(rootViewController: FeedListViewController(.init()))
}
#endif
